--- old/usr/src/cmd/svc/startd/graph.c	Mon Mar  6 21:30:20 2017
+++ new/usr/src/cmd/svc/startd/graph.c	Mon Mar  6 21:30:20 2017
@@ -23,6 +23,7 @@
  * Copyright (c) 2004, 2010, Oracle and/or its affiliates. All rights reserved.
  * Copyright (c) 2015, Syneto S.R.L. All rights reserved.
  * Copyright 2016 Toomas Soome <tsoome@me.com>
+ * Copyright 2016 RackTop Systems.
  */
 
 /*
@@ -208,9 +209,12 @@
 
 #define	is_depgrp_bypassed(v) ((v->gv_type == GVT_GROUP) && \
 	((v->gv_depgroup == DEPGRP_EXCLUDE_ALL) || \
-	(v->gv_depgroup == DEPGRP_OPTIONAL_ALL) || \
 	(v->gv_restart < RERR_RESTART)))
 
+#define	is_inst_bypassed(v) ((v->gv_type == GVT_INST) && \
+	((v->gv_flags & GV_TODISABLE) || \
+	(v->gv_flags & GV_TOOFFLINE)))
+
 static uu_list_pool_t *graph_edge_pool, *graph_vertex_pool;
 static uu_list_t *dgraph;
 static pthread_mutex_t dgraph_lock;
@@ -1298,7 +1302,7 @@
 			satisfiable = B_TRUE;
 	}
 
-	return (!satbility || satisfiable ? 0 : -1);
+	return ((!satbility || satisfiable) ? 0 : -1);
 }
 
 /*
@@ -1329,33 +1333,23 @@
 
 		switch (v->gv_type) {
 		case GVT_INST:
-			/* Skip missing or disabled instances */
-			if ((v->gv_flags & (GV_CONFIGURED | GV_ENABLED)) !=
-			    (GV_CONFIGURED | GV_ENABLED))
+			/* Skip missing instances */
+			if ((v->gv_flags & GV_CONFIGURED) == 0)
 				continue;
 
 			if (v->gv_state == RESTARTER_STATE_MAINT)
 				continue;
 
-			if (v->gv_flags & GV_TOOFFLINE)
-				continue;
-
 			any_qualified = B_TRUE;
-			if (v->gv_state == RESTARTER_STATE_OFFLINE) {
+			if (v->gv_state == RESTARTER_STATE_OFFLINE ||
+			    v->gv_state == RESTARTER_STATE_DISABLED) {
 				/*
-				 * For offline dependencies, treat unsatisfiable
-				 * as satisfied.
+				 * For offline/disabled dependencies,
+				 * treat unsatisfiable as satisfied.
 				 */
 				i = dependency_satisfied(v, B_TRUE);
 				if (i == -1)
 					i = 1;
-			} else if (v->gv_state == RESTARTER_STATE_DISABLED) {
-				/*
-				 * The service is enabled, but hasn't
-				 * transitioned out of disabled yet.  Treat it
-				 * as unsatisfied (not unsatisfiable).
-				 */
-				i = 0;
 			} else {
 				i = dependency_satisfied(v, satbility);
 			}
@@ -1368,68 +1362,9 @@
 			break;
 
 		case GVT_SVC: {
-			boolean_t svc_any_qualified;
-			boolean_t svc_satisfied;
-			boolean_t svc_satisfiable;
-			graph_vertex_t *v2;
-			graph_edge_t *e2;
-
-			svc_any_qualified = B_FALSE;
-			svc_satisfied = B_FALSE;
-			svc_satisfiable = B_FALSE;
-
-			for (e2 = uu_list_first(v->gv_dependencies);
-			    e2 != NULL;
-			    e2 = uu_list_next(v->gv_dependencies, e2)) {
-				v2 = e2->ge_vertex;
-				assert(v2->gv_type == GVT_INST);
-
-				if ((v2->gv_flags &
-				    (GV_CONFIGURED | GV_ENABLED)) !=
-				    (GV_CONFIGURED | GV_ENABLED))
-					continue;
-
-				if (v2->gv_state == RESTARTER_STATE_MAINT)
-					continue;
-
-				if (v2->gv_flags & GV_TOOFFLINE)
-					continue;
-
-				svc_any_qualified = B_TRUE;
-
-				if (v2->gv_state == RESTARTER_STATE_OFFLINE) {
-					/*
-					 * For offline dependencies, treat
-					 * unsatisfiable as satisfied.
-					 */
-					i = dependency_satisfied(v2, B_TRUE);
-					if (i == -1)
-						i = 1;
-				} else if (v2->gv_state ==
-				    RESTARTER_STATE_DISABLED) {
-					i = 0;
-				} else {
-					i = dependency_satisfied(v2, satbility);
-				}
-
-				if (i == 1) {
-					svc_satisfied = B_TRUE;
-					break;
-				}
-				if (i == 0)
-					svc_satisfiable = B_TRUE;
-			}
-
-			if (!svc_any_qualified)
-				continue;
 			any_qualified = B_TRUE;
-			if (svc_satisfied) {
-				i = 1;
-			} else if (svc_satisfiable) {
-				i = 0;
-			} else {
-				i = -1;
-			}
+			i = optional_all_satisfied(v, satbility);
+
 			break;
 		}
 
@@ -1609,24 +1544,35 @@
 		}
 
 		/*
-		 * Any vertex with the GV_TOOFFLINE flag set is guaranteed
-		 * to have its dependencies unsatisfiable.
+		 * Vertices may be transitioning so we try to figure out if
+		 * the end state is likely to satisfy the dependency instead
+		 * of assuming the dependency is unsatisfied/unsatisfiable.
+		 *
+		 * Support for optional_all dependencies depends on us getting
+		 * this right because unsatisfiable dependencies are treated
+		 * as being satisfied.
 		 */
-		if (v->gv_flags & GV_TOOFFLINE)
-			return (-1);
-
 		switch (v->gv_state) {
 		case RESTARTER_STATE_ONLINE:
 		case RESTARTER_STATE_DEGRADED:
+			if (v->gv_flags & GV_TODISABLE)
+				return (-1);
+			if (v->gv_flags & GV_TOOFFLINE)
+				return (0);
 			return (1);
 
 		case RESTARTER_STATE_OFFLINE:
-			if (!satbility)
-				return (0);
+			if (!satbility || v->gv_flags & GV_TODISABLE)
+				return (satbility ? -1 : 0);
 			return (instance_satisfied(v, satbility) != -1 ?
 			    0 : -1);
 
 		case RESTARTER_STATE_DISABLED:
+			if (!satbility || !(v->gv_flags & GV_ENABLED))
+				return (satbility ? -1 : 0);
+			return (instance_satisfied(v, satbility) != -1 ?
+			    0 : -1);
+
 		case RESTARTER_STATE_MAINT:
 			return (-1);
 
@@ -1729,6 +1675,9 @@
 static int
 satbility_cb(graph_vertex_t *v, void *arg)
 {
+	if (is_inst_bypassed(v))
+		return (UU_WALK_NEXT);
+
 	if (v->gv_type == GVT_INST)
 		graph_start_if_satisfied(v);
 
@@ -1743,13 +1692,34 @@
 
 static void propagate_stop(graph_vertex_t *, void *);
 
-/* ARGSUSED */
+/*
+ * propagate_start()
+ *
+ * This function is used to propagate a start event to the dependents of the
+ * given vertex.  Any dependents that are offline but have their dependencies
+ * satisfied are started.  Any dependents that are online and have restart_on
+ * set to "restart" or "refresh" are restarted because their dependencies have
+ * just changed.  This only happens with optional_all dependencies.
+ */
 static void
 propagate_start(graph_vertex_t *v, void *arg)
 {
+	restarter_error_t err = (restarter_error_t)arg;
+
+	if (is_inst_bypassed(v))
+		return;
+
 	switch (v->gv_type) {
 	case GVT_INST:
-		graph_start_if_satisfied(v);
+		/* Restarter */
+		if (inst_running(v)) {
+			if (err == RERR_RESTART || err == RERR_REFRESH) {
+				vertex_send_event(v,
+				    RESTARTER_EVENT_TYPE_STOP_RESET);
+			}
+		} else {
+			graph_start_if_satisfied(v);
+		}
 		break;
 
 	case GVT_GROUP:
@@ -1758,10 +1728,11 @@
 			    (void *)RERR_RESTART);
 			break;
 		}
+		err = v->gv_restart;
 		/* FALLTHROUGH */
 
 	case GVT_SVC:
-		graph_walk_dependents(v, propagate_start, NULL);
+		graph_walk_dependents(v, propagate_start, (void *)err);
 		break;
 
 	case GVT_FILE:
@@ -1781,13 +1752,23 @@
 	}
 }
 
+/*
+ * propagate_stop()
+ *
+ * This function is used to propagate a stop event to the dependents of the
+ * given vertex.  Any dependents that are online (or in degraded state) with
+ * the restart_on property set to "restart" or "refresh" will be stopped as
+ * their dependencies have just changed, propagate_start() will start them
+ * again once their dependencies have been re-satisfied.
+ */
 static void
 propagate_stop(graph_vertex_t *v, void *arg)
 {
-	graph_edge_t *e;
-	graph_vertex_t *svc;
 	restarter_error_t err = (restarter_error_t)arg;
 
+	if (is_inst_bypassed(v))
+		return;
+
 	switch (v->gv_type) {
 	case GVT_INST:
 		/* Restarter */
@@ -1815,7 +1796,8 @@
 
 	case GVT_GROUP:
 		if (v->gv_depgroup == DEPGRP_EXCLUDE_ALL) {
-			graph_walk_dependents(v, propagate_start, NULL);
+			graph_walk_dependents(v, propagate_start,
+			    (void *)RERR_NONE);
 			break;
 		}
 
@@ -1822,19 +1804,7 @@
 		if (err == RERR_NONE || err > v->gv_restart)
 			break;
 
-		assert(uu_list_numnodes(v->gv_dependents) == 1);
-		e = uu_list_first(v->gv_dependents);
-		svc = e->ge_vertex;
-
-		if (inst_running(svc)) {
-			if (err == RERR_RESTART || err == RERR_REFRESH) {
-				vertex_send_event(svc,
-				    RESTARTER_EVENT_TYPE_STOP_RESET);
-			} else {
-				vertex_send_event(svc,
-				    RESTARTER_EVENT_TYPE_STOP);
-			}
-		}
+		graph_walk_dependents(v, propagate_stop, arg);
 		break;
 
 	default:
@@ -4393,9 +4363,9 @@
 				return (B_FALSE);
 		} else {
 			/*
-			 * Skip all excluded and optional_all dependencies
-			 * and decide whether to offline the service based
-			 * on restart_on attribute.
+			 * Skip all excluded dependents and decide whether
+			 * to offline the service based on the restart_on
+			 * attribute.
 			 */
 			if (is_depgrp_bypassed(vv))
 				continue;
@@ -4960,7 +4930,7 @@
 	if (type == PROPAGATE_STOP) {
 		graph_walk_dependents(v, propagate_stop, (void *)rerr);
 	} else if (type == PROPAGATE_START || type == PROPAGATE_SAT) {
-		graph_walk_dependents(v, propagate_start, NULL);
+		graph_walk_dependents(v, propagate_start, (void *)RERR_NONE);
 
 		if (type == PROPAGATE_SAT)
 			propagate_satbility(v);
@@ -5031,7 +5001,7 @@
 	v->gv_flags &= ~GV_CONFIGURED;
 	v->gv_flags &= ~GV_DEATHROW;
 
-	graph_walk_dependents(v, propagate_start, NULL);
+	graph_walk_dependents(v, propagate_start, (void *)RERR_NONE);
 	propagate_satbility(v);
 
 	/*
@@ -5447,8 +5417,8 @@
 
 	switch (v->gv_type) {
 	case GVT_INST:
-		/* If the instance is already disabled, skip it. */
-		if (!(v->gv_flags & GV_ENABLED))
+		/* If the instance is already offline, skip it. */
+		if (!inst_running(v))
 			return (UU_WALK_NEXT);
 
 		v->gv_flags |= GV_TOOFFLINE;
@@ -5456,8 +5426,8 @@
 		break;
 	case GVT_GROUP:
 		/*
-		 * Skip all excluded and optional_all dependencies and decide
-		 * whether to offline the service based on restart_on attribute.
+		 * Skip all excluded dependents and decide whether to offline
+		 * the service based on the restart_on attribute.
 		 */
 		if (is_depgrp_bypassed(v))
 			return (UU_WALK_NEXT);
